from typing import Optional

from keycloak_admin_aio._lib.utils import (
    get_resource_id_in_location_header,
    remove_none,
)
from keycloak_admin_aio.types import ClientRepresentation

from .. import AttachedResources, KeycloakResource, KeycloakResourceWithIdentifierGetter
from .by_id import ClientsById


class Clients(KeycloakResource):
    """Provides the Keycloak clients resource.

    .. code:: python

        from keycloak_admin_aio import KeycloakAdmin, ClientRepresentation

        kc: KeycloakAdmin  # must be instantiated
    """

    _keycloak_resources: AttachedResources = [("by_id", ClientsById)]
    by_id: KeycloakResourceWithIdentifierGetter[ClientsById]

    def get_url(self) -> str:
        return f"{self._get_parent_url()}/clients"

    async def create(self, client_representation: ClientRepresentation) -> str:
        """Create a client.

        .. code:: python

            client_representation = ClientRepresentation(name="client-name")
            client_uuid: str = await kc.clients.create(client_representation)
        """
        connection = await self._get_connection()
        response = await connection.post(
            self.get_url(), json=client_representation.to_dict()
        )
        return get_resource_id_in_location_header(response)

    async def get(
        self,
        client_id: Optional[str] = None,
        first: Optional[int] = None,
        max: Optional[int] = None,
        q: Optional[str] = None,
        search: Optional[bool] = False,
        viewable_only: Optional[bool] = False,
    ) -> list[ClientRepresentation]:
        """Get clients.

        .. code:: python

            clients: list[ClientRepresentation] = await kc.clients.get()
        """
        connection = await self._get_connection()
        params = remove_none(
            {
                "clientId": client_id,
                "first": first,
                "max": max,
                "q": q,
                "search": search,
                "viewableOnly": viewable_only,
            }
        )
        response = await connection.get(self.get_url(), params=params)
        return ClientRepresentation.from_list(response.json())



from typing import Optional

from keycloak_admin_aio._lib.utils import (
    get_resource_id_in_location_header,
    remove_none,
)
from keycloak_admin_aio.types import GroupRepresentation

from .. import AttachedResources, KeycloakResource, KeycloakResourceWithIdentifierGetter
from .by_id import GroupsById


class Groups(KeycloakResource):
    """Provides the Keycloak group resource

    .. code:: python

        from keycloak_admin_aio import KeycloakAdmin, GroupRepresentation

        kc: KeycloakAdmin  # must be instantiated
    """

    _keycloak_resources: AttachedResources = [("by_id", GroupsById)]
    by_id: KeycloakResourceWithIdentifierGetter[GroupsById]

    def get_url(self) -> str:
        return f"{self._get_parent_url()}/groups"

    async def get(
        self,
        brief_representation: Optional[bool] = None,
        first: Optional[int] = None,
        max: Optional[int] = None,
        search: Optional[str] = None,
    ) -> list[GroupRepresentation]:
        """Get groups.

        .. code:: python

            groups: list[GroupRepresentation] = await kc.groups.get()
        """
        connection = await self._get_connection()
        params = remove_none(
            {
                "briefRepresentation": brief_representation,
                "first": first,
                "max": max,
                "search": search,
            }
        )
        response = await connection.get(self.get_url(), params=params)
        return GroupRepresentation.from_list(response.json())

    async def create(self, group_representation: GroupRepresentation) -> str:
        """Create a group.

        .. code:: python

            group_representation = GroupRepresentation(name="group-name")
            group_id: str = await kc.groups.create(group_representation)
        """
        connection = await self._get_connection()
        response = await connection.post(
            self.get_url(), json=group_representation.to_dict()
        )
        return get_resource_id_in_location_header(response)

    async def count(
        self, search: Optional[str] = None, top: Optional[bool] = None
    ) -> int:
        """Get group count.

        .. code:: python

            group_count: int = await kc.groups.count()
        """
        connection = await self._get_connection()
        params = remove_none({"search": search, "top": top})
        response = await connection.get(f"{self.get_url()}/count", params=params)
        return response.json()["count"]



from ..... import KeycloakResourceWithIdentifier


class UsersByIdGroupsById(KeycloakResourceWithIdentifier):
    def get_url(self) -> str:
        return f"{self._get_parent_url()}/{self.identifier}"

    async def create(self):
        connection = await self._get_connection()
        await connection.put(self.get_url())

    async def delete(self):
        connection = await self._get_connection()
        await connection.delete(self.get_url())



import httpx


async def raise_for_status_hook(response: httpx.Response):
    if response.status_code >= 400:
        await response.aread()
    response.raise_for_status()


httpx_default_args = {"event_hooks": {"response": [raise_for_status_hook]}}


def merge_with_default_httpx_args(user_args: dict) -> dict:
    return {**httpx_default_args, **user_args}



from __future__ import annotations

import abc
from typing import Any, Awaitable, Callable, Generic, TypeVar

import httpx

GetConnectionFn = Callable[..., Awaitable[httpx.AsyncClient]]
GetUrlFn = Callable[..., str]


def _create_getter(
    resource: Any,
    get_connection: GetConnectionFn,
    get_url: GetUrlFn,
):
    """Creates a getter function for identified resources.

    This function creator helper is required as using a lambda as an inline
    function leads to a reference problem with Python.
    """
    return lambda identifier: resource(get_connection, get_url, identifier)


class KeycloakResource:
    """Base class for all Keycloak resources.

    It attaches all the child resources in ``_keycloak_resources``
    """

    _keycloak_resources: AttachedResources

    def __init__(
        self,
        get_connection: GetConnectionFn,
        get_parent_url: GetUrlFn,
    ):
        """Initialize keycloak resource."""
        self._get_connection = get_connection
        self._get_parent_url = get_parent_url
        self.__set_keycloak_resources()

    @abc.abstractmethod
    def get_url(self) -> str:
        """Get the resource's url."""

    def __set_keycloak_resources(self):
        """Attaches all the specified child resources to the class instance.

        On init this method will parse the list of ``_keycloak_resources``
        attaching a getter for identified resources and a class instance for
        others.
        """
        if not hasattr(self, "_keycloak_resources"):
            return
        for resources_name, resource in self._keycloak_resources:
            if issubclass(resource, KeycloakResourceWithIdentifier):
                setattr(
                    self,
                    resources_name,
                    _create_getter(resource, self._get_connection, self.get_url),
                )
            else:
                setattr(
                    self, resources_name, resource(self._get_connection, self.get_url)
                )


class KeycloakResourceWithIdentifier(KeycloakResource):
    """Base class for all identified keycloak resources.

    These class instances will be provided via getters.
    E.g. ``kc.roles.by_name("some-role").get()``
    """

    def __init__(
        self,
        get_connection: GetConnectionFn,
        get_parent_url: GetUrlFn,
        identifier: str,
    ):
        """Initializes an identified keycloak resources adding the identifier."""
        super().__init__(get_connection, get_parent_url)
        self.identifier = identifier


AttachedResources = list[tuple[str, type[KeycloakResource]]]

T = TypeVar("T", bound=KeycloakResourceWithIdentifier)


class KeycloakResourceWithIdentifierGetter(Generic[T]):
    """Function which attaches to Keycloak resources to provide a child resource by identifier.

    Using ``Callable`` does not properly work and although (as far as I
    understand) it should be fixed: https://github.com/python/mypy/issues/708.
    Honestly I really don't understand what's going on here, but writing it
    this way resolved the mypy issue I was having
    """

    def __call__(self, identifier: str) -> T:
        ...



from keycloak_admin_aio.types import MappingsRepresentation

from .... import AttachedResources, KeycloakResource
from .realm import ClientScopesScopeMappingsRealm


class ClientScopesScopeMappings(KeycloakResource):
    """Scopes mappings for client scopes by id.

    .. code:: python

        from keycloak_admin_aio import KeycloakAdmin, MappingsRepresentation

        kc: KeycloakAdmin  # must be instantiated
        client_scope_id: str  # uuid
    """

    _keycloak_resources: AttachedResources = [("realm", ClientScopesScopeMappingsRealm)]
    realm: ClientScopesScopeMappingsRealm

    def get_url(self) -> str:
        return f"{self._get_parent_url()}/scope-mappings"

    async def get(self) -> MappingsRepresentation:
        """Get scope mappings for a client scope by id.

        .. code:: python

            client_scope_resource = kc.client_scopes.by_id(client_scope_id)
            mappings: MappingsRepresentation = await client_scope_resource.scope_mappings.get()
        """
        connection = await self._get_connection()
        response = await connection.get(self.get_url())
        return MappingsRepresentation.from_dict(response.json())



from keycloak_admin_aio.types import RequiredActionProviderRepresentation

from ... import KeycloakResource


class AuthenticationRequiredActions(KeycloakResource):
    """Required actions for a realm.

    .. code:: python

        from keycloak_admin_aio import KeycloakAdmin, RequiredActionProviderRepresentation

        kc: KeycloakAdmin  # must be instantiated
    """

    def get_url(self) -> str:
        return f"{self._get_parent_url()}/required-actions"

    async def get(self) -> list[RequiredActionProviderRepresentation]:
        """Get required actions.

        .. code:: python

            required_actions: list[
                RequiredActionProviderRepresentation
            ] = await kc.authentication.required_actions.get()
        """
        connection = await self._get_connection()
        response = await connection.get(self.get_url())
        return RequiredActionProviderRepresentation.from_list(response.json())



from typing import Optional

from keycloak_admin_aio._lib.utils import remove_none

from .... import KeycloakResource


class UsersByIdExecuteActionsEmail(KeycloakResource):
    """Provides the Keycloak user execute actions email resource.

    .. code:: python

        from keycloak_admin_aio import KeycloakAdmin

        kc: KeycloakAdmin  # needs to be instantiated
        user_id: str
    """

    def get_url(self) -> str:
        return f"{self._get_parent_url()}/execute-actions-email"

    async def send_email(
        self,
        actions: list[str],
        client_id: Optional[str] = None,
        lifespan: Optional[int] = None,
        redirect_uri: Optional[str] = None,
    ):
        """Send email to user.

        .. code:: python

            await kc.users.by_id(user_id).execute_actions_email.send_email()
        """
        connection = await self._get_connection()
        params = remove_none(
            {"client_id": client_id, "lifespan": lifespan, "redirect_uri": redirect_uri}
        )
        await connection.put(self.get_url(), params=params, json=actions)



from ..... import KeycloakResourceWithIdentifier


class ClientsByIdDefaultClientScopesById(KeycloakResourceWithIdentifier):
    """Client scopes by id for a client by UUID.

    .. code:: python

        from keycloak_admin_aio import KeycloakAdmin

        kc: KeycloakAdmin  # must be instantiated
        client_uuid: str
        client_scope_id: str  # uuid
    """

    def get_url(self) -> str:
        return f"{self._get_parent_url()}/{self.identifier}"

    async def create(self):
        """Add a client scope by id to a client by id.

        .. code:: python

            await kc.clients.by_id(client_uuid).default_client_scopes.by_id(client_scope_id).add()
        """
        connection = await self._get_connection()
        await connection.put(self.get_url())

    async def delete(self):
        """Remove a client scope by id from a client by id.

        .. code:: python

            await kc.clients.by_id(client_uuid).default_client_scopes.by_id(client_scope_id).delete()
        """
        connection = await self._get_connection()
        await connection.delete(self.get_url())



from ... import KeycloakResourceWithIdentifier


class SessionsById(KeycloakResourceWithIdentifier):
    """Realm sessions by id.

    .. code:: python

        from keycloak_admin_aio import KeycloakAdmin

        kc: KeycloakAdmin  # needs to be instantiated
        sesson_id: str  # uuid
    """

    def get_url(self) -> str:
        return f"{self._get_parent_url()}/{self.identifier}"

    async def delete(self):
        """Terminate a session by id.

        .. code:: python

            await kc.sessions.by_id(sesson_id).delete()
        """
        connection = await self._get_connection()
        await connection.delete(self.get_url())



from keycloak_admin_aio.types import GroupRepresentation

from .... import KeycloakResource


class GroupsByIdChildren(KeycloakResource):
    """Get children of a group by id.

    .. code:: python

        from keycloak_admin_aio import KeycloakAdmin, GroupRepresentation

        kc: KeycloakAdmin  # must be instantiated
        group_id: str  # uuid
    """

    def get_url(self) -> str:
        return f"{self._get_parent_url()}/children"

    async def create(self, group_representation: GroupRepresentation):
        """Create a parent/child group relationship.

        .. code:: python

            group_representation = GroupRepresentation(name="child-group-name")
            await kc.goups.by_id(group_id).children.add(group_representation)
        """
        connection = await self._get_connection()
        await connection.post(self.get_url(), json=group_representation.to_dict())



from keycloak_admin_aio.types import ClientRepresentation

from ... import AttachedResources, KeycloakResourceWithIdentifier
from .default_client_scopes import ClientsByIdDefaultClientScopes
from .user_sessions import ClientsByIdUserSessions


class ClientsById(KeycloakResourceWithIdentifier):
    """Clients by UUID.

    .. code:: python

        from keycloak_admin_aio import KeycloakAdmin, ClientRepresentation

        kc: KeycloakAdmin  # must be instantiated
        client_uuid: str
    """

    _keycloak_resources: AttachedResources = [
        ("default_client_scopes", ClientsByIdDefaultClientScopes),
        ("user_sessions", ClientsByIdUserSessions),
    ]
    default_client_scopes: ClientsByIdDefaultClientScopes
    user_sessions: ClientsByIdUserSessions

    def get_url(self) -> str:
        return f"{self._get_parent_url()}/{self.identifier}"

    async def get(self) -> ClientRepresentation:
        """Get a client by UUID.

        .. code:: python

            client_representation: ClientRepresentation = await kc.clients.by_id(client_uuid).get()
        """
        connection = await self._get_connection()
        response = await connection.get(self.get_url())
        return ClientRepresentation.from_dict(response.json())

    async def update(self, client_representation: ClientRepresentation):
        """Update a client by UUID.

        .. code:: python

            client_representation = ClientRepresentation(name="client-name")
            await kc.clients.by_id(client_uuid).update(client_representation)
        """
        connection = await self._get_connection()
        await connection.put(self.get_url(), json=client_representation.to_dict())

    async def delete(self):
        """Delete a client by UUID.

        .. code:: python

            await kc.clients.by_id(client_uuid).delete()
        """
        connection = await self._get_connection()
        await connection.delete(self.get_url())



from keycloak_admin_aio.types import RoleRepresentation

from .... import KeycloakResource


class RolesByIdComposites(KeycloakResource):
    """Get composites for a role by id.

    .. code:: python

        from keycloak_admin_aio import KeycloakAdmin, RoleRepresentation

        kc: KeycloakAdmin  # needs to be instantiated
        role_id: str  # uuid
    """

    def get_url(self) -> str:
        return f"{self._get_parent_url()}/composites"

    async def create(self, composite_roles: list[RoleRepresentation]):
        """Create composites for a role by id.

        .. code:: python

            role_representations: list[RoleRepresentation] = []  # needs to be populated
            await kc.roles.by_id(role_id).composites.create(role_representations)
        """
        connection = await self._get_connection()
        await connection.post(
            self.get_url(),
            json=RoleRepresentation.to_dict_list(composite_roles),
        )

    async def get(self) -> list[RoleRepresentation]:
        """Get composites for a role by id.

        .. code:: python

            composites: list[RoleRepresentation] = await kc.roles.by_id(role_id).composites.get()
        """
        connection = await self._get_connection()
        response = await connection.get(self.get_url())
        response.raise_for_status()
        return RoleRepresentation.from_list(response.json())

    async def delete(self, composite_roles: list[RoleRepresentation]):
        """Delete composites for a role by id.

        .. code:: python

            role_representations: list[RoleRepresentation] = []  # needs to be populated
            await kc.roles.by_id(role_id).composites.delete(role_representations)
        """
        connection = await self._get_connection()
        await connection.request(
            method="DELETE",
            url=self.get_url(),
            json=RoleRepresentation.to_dict_list(composite_roles),
        )



from keycloak_admin_aio.types import RoleRepresentation

from ..... import KeycloakResource


class UsersByIdRoleMappingsRealm(KeycloakResource):
    """Provides the realm role mappings for users.

    .. code:: python

        from keycloak_admin_aio import KeycloakAdmin, RoleRepresentation

        kc: KeycloakAdmin  # needs to be instantiated
        user_id: str  # uuid
    """

    def get_url(self) -> str:
        return f"{self._get_parent_url()}/realm"

    async def get(self) -> list[RoleRepresentation]:
        """Get realm role mappings for a user.

        .. code:: python

            roles: list[RoleRepresentation] = await kc.users.by_id(user_id).role_mappings.realm.get()
        """
        connection = await self._get_connection()
        response = await connection.get(self.get_url())
        return RoleRepresentation.from_list(response.json())

    async def create(self, role_representations: list[RoleRepresentation]):
        """Create realm role mappings for a user.

        .. code:: python

            roles list[RoleRepresentation] = []  # needs to be populated
            await kc.users.by_id(user_id).role_mappings.realm.create(roles)
        """
        connection = await self._get_connection()
        await connection.post(
            self.get_url(),
            json=RoleRepresentation.to_dict_list(role_representations),
        )

    async def delete(self, role_representations: list[RoleRepresentation]):
        """Delete realm roles mappings for a user.

        .. code:: python

            roles: list[RoleRepresentation] = []  # needs to be populated
            await kc.users.by_id(user_id).role_mappings.realm.delete(roles)
        """
        connection = await self._get_connection()
        await connection.request(
            method="DELETE",
            url=self.get_url(),
            json=RoleRepresentation.to_dict_list(role_representations),
        )

    async def available(self) -> list[RoleRepresentation]:
        """Get available roles which can be mapped to the user.

        .. code:: python

            roles: list[RoleRepresentation] = await kc.users.by_id(user_id).role_mappings.realm.available()
        """
        connection = await self._get_connection()
        response = await connection.get(f"{self.get_url()}/available")
        return RoleRepresentation.from_list(response.json())

    async def composite(self) -> list[RoleRepresentation]:
        """Get composed realm roles for a user.

        .. code:: python

            roles: list[RoleRepresentation] = await kc.users.by_id(user_id).role_mappings.realm.composite()
        """

        connection = await self._get_connection()
        response = await connection.get(f"{self.get_url()}/composite")
        return RoleRepresentation.from_list(response.json())



from .. import AttachedResources, KeycloakResource, KeycloakResourceWithIdentifierGetter
from .by_id import SessionsById


class Sessions(KeycloakResource):
    """Provides realm sessions."""

    _keycloak_resources: AttachedResources = [("by_id", SessionsById)]
    by_id: KeycloakResourceWithIdentifierGetter[SessionsById]

    def get_url(self) -> str:
        return f"{self._get_parent_url()}/sessions"



from keycloak_admin_aio.types import RoleRepresentation

from ..... import KeycloakResource


class ClientScopesScopeMappingsRealm(KeycloakResource):
    """Realm scope mappings for a client scope by id.

    .. code:: python

        from keycloak_admin_aio import KeycloakAdmin, RoleRepresentation

        kc: KeycloakAdmin  # must be instantiated
        client_scope_id: str  # uuid
    """

    def get_url(self) -> str:
        return f"{self._get_parent_url()}/realm"

    async def create(self, role_representations: list[RoleRepresentation]):
        """Add roles to the realm scope mappings of a client scope by id.

        .. code:: python

            role_representations: list[RoleRepresentation] = []  # needs to be populated
            await kc.client_scopes.by_id(client_scope_id).scope_mappings.realm.create(roles)
        """
        connection = await self._get_connection()
        await connection.post(
            self.get_url(),
            json=RoleRepresentation.to_dict_list(role_representations),
        )

    async def get(self) -> list[RoleRepresentation]:
        """Get realm scope mappings for a client scope by id.

        .. code:: python

            client_scope_resource = kc.client_scopes.by_id(client_scope_id)
            roles: list[RoleRepresentation] = await client_scope_resource.scope_mappings.realm.get()
        """
        connection = await self._get_connection()
        response = await connection.get(self.get_url())
        return RoleRepresentation.from_list(response.json())

    async def delete(self, role_representations: list[RoleRepresentation]):
        """Remove roles from realm scope mappings for a client scope by id.

        .. code:: python

            role_representations: list[RoleRepresentation] = []  # needs to be populated
            client_scope_resource = kc.client_scopes.by_id(client_scope_id)
            await client_scope_resource.scope_mappings.realm.delete(role_representations)
        """
        connection = await self._get_connection()
        await connection.request(
            method="DELETE",
            url=self.get_url(),
            json=RoleRepresentation.to_dict_list(role_representations),
        )



from keycloak_admin_aio.types import ClientScopeRepresentation

from ... import AttachedResources, KeycloakResourceWithIdentifier
from .scope_mappings import ClientScopesScopeMappings


class ClientScopesById(KeycloakResourceWithIdentifier):
    """Client scopes by id.

    .. code:: python

        from keycloak_admin_aio import KeycloakAdmin, ClientScopeRepresentation

        kc: KeycloakAdmin  # must be instantiated
        client_scope_id: str  # uuid
    """

    _keycloak_resources: AttachedResources = [
        ("scope_mappings", ClientScopesScopeMappings)
    ]
    scope_mappings: ClientScopesScopeMappings

    def get_url(self) -> str:
        return f"{self._get_parent_url()}/{self.identifier}"

    async def get(self) -> ClientScopeRepresentation:
        """Get a client scope by id.

        .. code:: python

            client_scope: ClientScopeRepresentation = await kc.client_scopes.by_id(client_scope_id).get()
        """
        connection = await self._get_connection()
        response = await connection.get(self.get_url())
        return ClientScopeRepresentation.from_dict(response.json())

    async def update(
        self,
        client_scope_representation: ClientScopeRepresentation,
    ):
        """Update a client scope by id.

        .. code:: python

            client_scope_representation = ClientScopeRepresentation(name="client-scope-name")
            await kc.client_scopes.by_id(client_scope_id).update(client_scope_representation)
        """
        connection = await self._get_connection()
        await connection.put(self.get_url(), json=client_scope_representation.to_dict())

    async def delete(self):
        """Delete a client scope by id.

        .. code:: python

            await kc.client_scopes.by_id(client_scope_id).delete()
        """
        connection = await self._get_connection()
        await connection.delete(self.get_url())



"""
This module provides a typed schema of the Keycloak input and outputs.

There's a few ways in which you can instantiate a dataclass including some not
provided by the standard libarary:

Instantiate a dataclass with keyword arguments
----------------------------------------------

.. code:: python

    from keycloak_admin_aio import RoleRepresentation

    role_representation = RoleRepresentation(
        name="role-name", description="role-description"
    )

    ## or from a dict

    role_representation = RoleRepresentation(
        **{"name": "role-name", "description": "role-description"}
    )


Instantiate a dataclass with the from_dict class method
-------------------------------------------------------

The above works well if you have a flat data model. If you have a dataclass
which uses other dataclasses using ``**`` with a dict will not work as
expected.  If you have a nested data structure you should use the ``from_dict``
function.

.. code:: python

    from keycloak_admin_aio import ClientMappingsRepresentation

    client_mapping_representation = ClientMappingsRepresentation(
        client="some-client", mappings=[RoleRepresentation(name="some-role")]
    )

    # can also be done like this

    client_mapping_representation = ClientMappingsRepresentation.from_dict(
        {"client": "some-client", "mappings": [{"name": "some-role"}]}
    )

    assert client_mapping_representation.mappings[0].name == "some-role"

Instantiate a list of dataclasses with the from_list class method
-----------------------------------------------------------------

If you have a list of dictionaries you want to parse into a list of dataclasses
you use the ``from_list`` method.

.. code:: python

    from keycloak_admin_aio import RoleRepresentation

    roles = [{"name": "A"}, {"name": "B"}, {"name": "C"}]
    roles = RoleRepresentation.from_list(roles)
"""

# Needed for sphinx autodoc
__all__ = [
    "RoleRepresentation",
    "RoleRepresentationComposites",
    "ClientMappingsRepresentation",
    "MappingsRepresentation",
    "ProtocolMapperRepresentation",
    "ClientScopeRepresentation",
    "FederatedIdentityRepresentation",
    "CredentialRepresentation",
    "UserConsentRepresentation",
    "UserRepresentation",
    "GroupRepresentation",
    "ScopeRepresentation",
    "ResourceRepresentation",
    "PolicyRepresentation",
    "ResourceServerRepresentation",
    "ClientRepresentation",
    "RequiredActionProviderRepresentation",
    "UserSession",
]

from .types import *



from typing import Optional

from keycloak_admin_aio._lib.utils import (
    get_resource_id_in_location_header,
    remove_none,
)
from keycloak_admin_aio.types import RoleRepresentation

from .. import AttachedResources, KeycloakResource, KeycloakResourceWithIdentifierGetter
from .by_id import RolesById
from .by_name import RolesByName


class Roles(KeycloakResource):
    """Provides the Keycloak role resource.

    .. code:: python

        from keycloak_admin_aio import KeycloakAdmin, UserRepresentation

        kc: KeycloakAdmin  # needs to be instantiated
    """

    _keycloak_resources: AttachedResources = [
        ("by_id", RolesById),
        ("by_name", RolesByName),
    ]
    by_name: KeycloakResourceWithIdentifierGetter[RolesByName]
    by_id: KeycloakResourceWithIdentifierGetter[RolesById]

    def get_url(self) -> str:
        return f"{self._get_parent_url()}/roles"

    async def create(self, role_representation: RoleRepresentation) -> str:
        """Create a role.

        .. code:: python

            role_representation = RoleRepresentation(name="some-role")
            role_id: str = await kc.roles.create(role_representation)
        """
        connection = await self._get_connection()
        response = await connection.post(
            self.get_url(), json=role_representation.to_dict()
        )
        role_name = get_resource_id_in_location_header(response, is_no_uuid=True)
        return role_name

    async def get(
        self,
        brief_representation: Optional[bool] = None,
        first: Optional[int] = None,
        max: Optional[int] = None,
        search: Optional[str] = None,
    ) -> list[RoleRepresentation]:
        """Get roles.

        .. code:: python

            roles: list[RoleRepresentation] = kc.roles.get()
        """
        connection = await self._get_connection()
        params = remove_none(
            {
                "briefRepresentation": brief_representation,
                "first": first,
                "search": search,
                "max": max,
            }
        )
        response = await connection.get(
            self.get_url(),
            params=params,
        )
        return RoleRepresentation.from_list(response.json())



from keycloak_admin_aio._lib.utils import get_resource_id_in_location_header
from keycloak_admin_aio.types import ClientScopeRepresentation

from .. import AttachedResources, KeycloakResource, KeycloakResourceWithIdentifierGetter
from .by_id import ClientScopesById


class ClientScopes(KeycloakResource):
    """Provides the Keycloak client scopes resource.

    .. code:: python

        from keycloak_admin_aio import KeycloakAdmin, ClientScopeRepresentation

        kc: KeycloakAdmin  # must be instantiated
    """

    _keycloak_resources: AttachedResources = [("by_id", ClientScopesById)]
    by_id: KeycloakResourceWithIdentifierGetter[ClientScopesById]

    def get_url(self) -> str:
        return f"{self._get_parent_url()}/client-scopes"

    async def create(
        self, client_scope_representation: ClientScopeRepresentation
    ) -> str:
        """Create a client scope.

        .. code:: python

            client_scope_representation = ClientScopeRepresentation(name="client-scope-name")
            client_scope_id: str = await kc.client_scopes.create(client_scope_representation)
        """
        connection = await self._get_connection()
        response = await connection.post(
            self.get_url(), json=client_scope_representation.to_dict()
        )
        return get_resource_id_in_location_header(response)

    async def get(self):
        """Get client scopes.

        .. code:: python

            client_scopes: list[ClientScopeRepresentation] = await kc.client_scopes.get()
        """
        connection = await self._get_connection()
        response = await connection.get(self.get_url())
        return ClientScopeRepresentation.from_list(response.json())



from .. import AttachedResources, KeycloakResource
from .required_actions import AuthenticationRequiredActions


class Authentication(KeycloakResource):
    """Provides the Keycloak authentication management resource."""

    _keycloak_resources: AttachedResources = [
        ("required_actions", AuthenticationRequiredActions)
    ]
    required_actions: AuthenticationRequiredActions

    def get_url(self) -> str:
        return f"{self._get_parent_url()}/authentication"



from typing import Optional

from keycloak_admin_aio._lib.utils import remove_none
from keycloak_admin_aio.types import OperationType, ResourceType

from .. import KeycloakResource


class AdminEvents(KeycloakResource):
    """Provides the Keycloak admin events resource.

    .. code:: python

        from keycloak_admin_aio import KeycloakAdmin

        kc: KeycloakAdmin  # must be instantiated
    """

    def get_url(self) -> str:
        return f"{self._get_parent_url()}/admin-events"

    async def get(
        self,
        auth_client: Optional[str] = None,
        auth_ip_address: Optional[str] = None,
        auth_realm: Optional[str] = None,
        auth_user: Optional[str] = None,
        date_from: Optional[str] = None,
        date_to: Optional[str] = None,
        first: Optional[int] = None,
        max: Optional[int] = None,
        operation_types: Optional[list[OperationType]] = None,
        resource_path: Optional[str] = None,
        resource_types: Optional[list[ResourceType]] = None,
    ) -> list[dict]:
        """Get admin events.

        .. code:: python

            events: list[dict] = await kc.admin_events.get()
        """
        connection = await self._get_connection()
        params = remove_none(
            {
                "authClient": auth_client,
                "authIpAddress": auth_ip_address,
                "authRealm": auth_realm,
                "authUser": auth_user,
                "dateFrom": date_from,
                "dateTo": date_to,
                "first": first,
                "max": max,
                "operationTypes": operation_types,
                "resourcePath": resource_path,
                "resourceTypes": resource_types,
            }
        )
        response = await connection.get(self.get_url(), params=params)
        return response.json()



from __future__ import annotations

from typing import Any, TypeVar

from dacite.core import from_dict

from keycloak_admin_aio._lib.utils import asdict


class DataClass:
    def to_dict(self) -> dict[str, Any]:
        return asdict(self)

    @staticmethod
    def to_dict_list(representation_list: list[T]) -> list[dict[str, Any]]:
        return list(
            map(lambda representation: representation.to_dict(), representation_list)
        )

    @classmethod
    def from_dict(cls, dictionary: dict):
        return from_dict(data_class=cls, data=dictionary)

    @classmethod
    def from_list(cls, _list: list[dict]):
        return list(map(cls.from_dict, _list))


T = TypeVar("T", bound=DataClass)



from typing import Optional

from keycloak_admin_aio._lib.utils import (
    get_resource_id_in_location_header,
    remove_none,
)
from keycloak_admin_aio.types import UserRepresentation

from .. import AttachedResources, KeycloakResource, KeycloakResourceWithIdentifierGetter
from .by_id import UsersById


class Users(KeycloakResource):
    """Provides the Keycloak user resource.

    .. code:: python

        from keycloak_admin_aio import KeycloakAdmin, UserRepresentation

        kc: KeycloakAdmin  # needs to be instantiated
    """

    _keycloak_resources: AttachedResources = [("by_id", UsersById)]
    by_id: KeycloakResourceWithIdentifierGetter[UsersById]

    def get_url(self) -> str:
        return f"{self._get_parent_url()}/users"

    async def get(
        self,
        brief_representation: Optional[bool] = None,
        first: Optional[int] = None,
        max: Optional[int] = None,
        search: Optional[str] = None,
        email: Optional[str] = None,
        email_verified: Optional[bool] = None,
        enabled: Optional[bool] = None,
        exact: Optional[bool] = None,
        first_name: Optional[str] = None,
        last_name: Optional[str] = None,
        idp_alias: Optional[str] = None,
        idp_user_id: Optional[str] = None,
        username: Optional[str] = None,
    ) -> list[UserRepresentation]:
        """Get users.

        .. code:: python

            users: list[UserRepresentation] = await kc.users.get()
        """
        connection = await self._get_connection()
        params = remove_none(
            {
                "briefRepresentation": brief_representation,
                "first": first,
                "max": max,
                "search": search,
                "email": email,
                "emailVerified": email_verified,
                "enabled": enabled,
                "exact": exact,
                "firstName": first_name,
                "lastName": last_name,
                "idpAlias": idp_alias,
                "ipdUserId": idp_user_id,
                "username": username,
            }
        )
        response = await connection.get(self.get_url(), params=params)
        return UserRepresentation.from_list(response.json())

    async def create(self, user_representation: UserRepresentation) -> str:
        """Create user.

        .. code:: python

            user_representation = UserRepresentation(email="user@domain.com")
            user_id: str = await kc.users.create(user_representation)
        """
        connection = await self._get_connection()
        response = await connection.post(
            self.get_url(), json=user_representation.to_dict()
        )
        return get_resource_id_in_location_header(response)

    async def count(
        self,
        search: Optional[str] = None,
        email: Optional[str] = None,
        email_verified: Optional[bool] = None,
        exact: Optional[bool] = None,
        first_name: Optional[str] = None,
        last_name: Optional[str] = None,
        username: Optional[str] = None,
    ) -> int:
        """User count.

        .. code:: python

            user_count: int = await kc.users.count()
        """
        connection = await self._get_connection()
        params = remove_none(
            {
                "search": search,
                "email": email,
                "emailVerified": email_verified,
                "exact": exact,
                "firstName": first_name,
                "lastName": last_name,
                "username": username,
            }
        )
        response = await connection.get(f"{self.get_url()}/count", params=params)
        return int(response.text)



from keycloak_admin_aio.types import RoleRepresentation

from ... import AttachedResources, KeycloakResourceWithIdentifier
from .composites import RolesByIdComposites


class RolesById(KeycloakResourceWithIdentifier):
    """Provides roles by id.

    .. code:: python

        from keycloak_admin_aio import KeycloakAdmin, RoleRepresentation

        kc: KeycloakAdmin  # needs to be instantiated
        role_id: str  # uuid
    """

    _keycloak_resources: AttachedResources = [("composites", RolesByIdComposites)]
    composites: RolesByIdComposites

    def get_url(self) -> str:
        return f"{self._get_parent_url()}-by-id/{self.identifier}"

    async def get(self) -> RoleRepresentation:
        """Get a role by id.

        .. code:: python

            role: RoleRepresentation = await kc.roles.by_id(role_id).get()
        """
        connection = await self._get_connection()
        response = await connection.get(self.get_url())
        return RoleRepresentation.from_dict(response.json())

    async def update(self, role_representation: RoleRepresentation):
        """Update a role by id.

        .. code:: python

            role_representation = RoleRepresentation(name="role-name")
            await kc.roles.by_id(role_id).update(role_representation)
        """
        connection = await self._get_connection()
        await connection.put(self.get_url(), json=role_representation.to_dict())

    async def delete(self):
        """Delete a role by id.

        .. code:: python

            await kc.roles.by_id(role_id).delete()
        """
        connection = await self._get_connection()
        await connection.delete(self.get_url())



from typing import Optional

from keycloak_admin_aio._lib.utils import remove_none
from keycloak_admin_aio.types import UserRepresentation

from .... import KeycloakResource


class GroupsByIdMembers(KeycloakResource):
    """Members of groups by id.

    .. code:: python

        from keycloak_admin_aio import KeycloakAdmin, UserRepresentation

        kc: KeycloakAdmin  # must be instantiated
        group_id: str  # uuid
    """

    def get_url(self) -> str:
        return f"{self._get_parent_url()}/members"

    async def get(
        self,
        brief_representation: Optional[bool] = None,
        first: Optional[int] = None,
        max: Optional[int] = None,
    ) -> list[UserRepresentation]:
        """Get members of a group by id.

        .. code:: python

            members: list[UserRepresentation] = await kc.groups.by_id(group_id).members.get()
        """
        connection = await self._get_connection()
        params = remove_none(
            {
                "briefRepresentation": brief_representation,
                "first": first,
                "max": max,
            }
        )
        response = await connection.get(self.get_url(), params=params)
        return UserRepresentation.from_list(response.json())



from typing import Optional

from keycloak_admin_aio._lib.utils import remove_none
from keycloak_admin_aio.types import GroupRepresentation

from .... import (
    AttachedResources,
    KeycloakResource,
    KeycloakResourceWithIdentifierGetter,
)
from .by_id import UsersByIdGroupsById


class UsersByIdGroups(KeycloakResource):
    """Provides the Keycloak groups resource for users by id

    .. code:: python

        from keycloak_admin_aio import KeycloakAdmin, GroupRepresentation

        kc: KeycloakAdmin  # needs to be instantiated
        user_id: str  # uuid
    """

    _keycloak_resources: AttachedResources = [("by_id", UsersByIdGroupsById)]
    by_id: KeycloakResourceWithIdentifierGetter[UsersByIdGroupsById]

    def get_url(self) -> str:
        return f"{self._get_parent_url()}/groups"

    async def get(
        self,
        first: Optional[int] = None,
        max: Optional[int] = None,
        search: Optional[str] = None,
        brief_representation: bool = True,
    ) -> list[GroupRepresentation]:
        """Get the user's groups.

        .. code:: python

            groups: list[GroupRepresentation] = await kc.users.by_id(user_id).groups.get()
        """
        connection = await self._get_connection()
        params = remove_none(
            {
                "first": first,
                "max": max,
                "search": search,
                "briefRepresentation": brief_representation,
            }
        )
        response = await connection.get(self.get_url(), params=params)
        return GroupRepresentation.from_list(response.json())

    async def count(
        self,
        search: Optional[str] = None,
    ) -> int:
        """Get the number of groups the user belongs to.

        .. code:: python

            membership_count: int = await kc.users.by_id(user_id).groups.count()
        """
        connection = await self._get_connection()
        params = remove_none({"search": search})
        response = await connection.get(self.get_url(), params=params)
        return int(response.json()["count"])



from keycloak_admin_aio.types import MappingsRepresentation

from .... import AttachedResources, KeycloakResource
from .realm import UsersByIdRoleMappingsRealm


class UsersByIdRoleMappings(KeycloakResource):
    """Provides the role mappings for a user.

    .. code:: python

        from keycloak_admin_aio import KeycloakAdmin, MappingsRepresentation

        kc: KeycloakAdmin  # needs to be instantiated
        user_id: str  # uuid
    """

    _keycloak_resources: AttachedResources = [("realm", UsersByIdRoleMappingsRealm)]
    realm: UsersByIdRoleMappingsRealm

    def get_url(self) -> str:
        return f"{self._get_parent_url()}/role-mappings"

    async def get(self) -> MappingsRepresentation:
        """Get the role mappings for a user.

        .. code:: python

            mappings: MappingsRepresentation = await kc.users.by_id(user_id).role_mappings.get()
        """
        connection = await self._get_connection()
        response = await connection.get(self.get_url())
        return MappingsRepresentation.from_dict(response.json())



from __future__ import annotations

import asyncio
from datetime import datetime, timedelta
from typing import Any, Literal, Optional

import httpx

from ._httpx_args import merge_with_default_httpx_args
from ._lib.utils import remove_none
from ._resources import (
    AdminEvents,
    AttachedResources,
    Authentication,
    Clients,
    ClientScopes,
    Groups,
    Roles,
    Sessions,
    Users,
)


class KeycloakAdmin:
    """Base class for Keycloak Admin API endpoints.

    It handles the ``access_token`` and guarantees it being valid when using the
    ``get_access_token`` method or accessing a protected Keycloak resource.
    """

    __keycloak_resources: AttachedResources = [
        ("roles", Roles),
        ("client_scopes", ClientScopes),
        ("users", Users),
        ("clients", Clients),
        ("admin_events", AdminEvents),
        ("authentication", Authentication),
        ("groups", Groups),
        ("sessions", Sessions),
    ]
    roles: Roles
    """https://www.keycloak.org/docs-api/15.0/rest-api/index.html#_roles_resource"""

    client_scopes: ClientScopes
    """https://www.keycloak.org/docs-api/15.0/rest-api/index.html#_client_scopes_resource"""

    users: Users
    """https://www.keycloak.org/docs-api/15.0/rest-api/index.html#_users_resource"""

    clients: Clients
    """https://www.keycloak.org/docs-api/15.0/rest-api/index.html#_clients_resource"""

    admin_events: AdminEvents
    """https://www.keycloak.org/docs-api/15.0/rest-api/index.html#_realms_admin_resource"""

    authentication: Authentication
    """https://www.keycloak.org/docs-api/15.0/rest-api/index.html#_authentication_management_resource"""

    groups: Groups
    """https://www.keycloak.org/docs-api/15.0/rest-api/index.html#_groups_resource"""

    sessions: Sessions

    leeway: int
    """A token will be considered as expired seconds before its actual expiry controlled by this value."""

    def __init__(
        self,
        server_url: str,
        grant_type: Literal["client_credentials", "password"],
        username: Optional[str] = None,
        password: Optional[str] = None,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        realm: str = "master",
        leeway: int = 10,
        httpx_args={},
    ):
        """Initialize ``KeycloakAdmin`` with either client or user credentials.

        Should not be used directly. The usage
        of ``with_client_credentials`` or ``with_password`` should be
        preferred.
        """
        allowed_grant_types = ["client_credentials", "password"]
        if not any([grant_type == allowed for allowed in allowed_grant_types]):
            raise Exception(f"'grant_type' needs to be in '{allowed_grant_types}'")
        self._realm = realm
        self._username = username
        self._client_id = client_id
        self._client_secret = client_secret
        self._password = password
        self._server_url = server_url
        self._grant_type = grant_type
        self.leeway = leeway
        self.__connection = httpx.AsyncClient(
            **merge_with_default_httpx_args(httpx_args)
        )
        self.__access_token = None
        self.__refresh_token = None
        self.__set_keycloak_resources()
        self.__lock = asyncio.Lock()

    @classmethod
    def with_client_credentials(
        cls,
        server_url: str,
        client_id: str,
        client_secret: str,
        realm: str = "master",
        leeway: int = 10,
        httpx_args={},
    ) -> KeycloakAdmin:
        """Instantiate ``KeycloakAdmin`` with ``client_id`` and ``client_secret``."""
        return cls(
            server_url,
            grant_type="client_credentials",
            client_id=client_id,
            client_secret=client_secret,
            realm=realm,
            leeway=leeway,
            httpx_args=httpx_args,
        )

    @classmethod
    def with_password(
        cls,
        server_url: str,
        username: Optional[str] = None,
        password: Optional[str] = None,
        client_id: Optional[str] = "admin-cli",
        realm: str = "master",
        leeway: int = 10,
        httpx_args={},
    ) -> KeycloakAdmin:
        """Instantiate ``KeycloakAdmin`` with user credentials (username and password)."""
        return cls(
            server_url,
            grant_type="password",
            client_id=client_id,
            username=username,
            password=password,
            realm=realm,
            leeway=leeway,
            httpx_args=httpx_args,
        )

    @property
    def realm(self):
        """Realm property."""
        return self._realm

    @property
    def username(self):
        """Username property."""
        return self._username

    @property
    def client_id(self):
        """Client id property."""
        return self._client_id

    @property
    def client_secret(self):
        """Client secret property."""
        return self.client_secret

    @property
    def password(self):
        """Password property."""
        return self._password

    @property
    def server_url(self):
        """Server url property."""
        return self._server_url

    @property
    def grant_type(self):
        """Token endpoint grant type."""
        return self._grant_type

    def get_url(self):
        """Get the admin api base url."""
        return f"{self._server_url}/admin/realms/{self._realm}"

    async def close(self):
        """Closes open httpx connection."""
        await self.__connection.aclose()

    def close_sync(self):
        """Synchronously close open httpx connection."""
        loop = asyncio.get_event_loop()
        loop.run_until_complete(self.close())

    def __set_keycloak_resources(self):
        for resources_name, resource in self.__keycloak_resources:
            setattr(
                self,
                resources_name,
                resource(self.__get_connection, self.get_url),
            )

    async def __aenter__(self) -> KeycloakAdmin:
        """For entering asynchronous context manger."""
        return self

    async def __aexit__(self, *_, **__):
        """Cleanup for asynchronous context manger."""
        await self.close()

    async def get_access_token(self):
        """Get ``access_token``. Guaranteed not to be expired."""
        async with self.__lock:
            if not self.__access_token:
                await self.__token()
            else:
                now = datetime.now()
                if now > self.access_token_expire:
                    if now < self.refresh_token_expire:
                        await self.__token_refresh()
                    else:
                        await self.__token()
            return self.__access_token

    def get_token_url(self) -> str:
        """Openid connect token endpoint url."""
        return f"{self._server_url}/realms/{self._realm}/protocol/openid-connect/token"

    def __parse_token_response(self, token_response: dict[str, Any]):
        self.__access_token = token_response["access_token"]
        self.__refresh_token = token_response.get("refresh_token")
        self.access_token_expire = datetime.now() + timedelta(
            seconds=token_response["expires_in"] - self.leeway
        )
        self.refresh_token_expire = datetime.now() + timedelta(
            seconds=token_response.get("refresh_expires_in", 0) - self.leeway
        )

    async def __token_refresh(self):
        headers = httpx.Headers({"Content-Type": "application/x-www-form-urlencoded"})
        payload = remove_none(
            {
                "client_id": self._client_id,
                "client_secret": self._client_secret,
                "grant_type": "refresh_token",
                "refresh_token": self.__refresh_token,
            }
        )
        try:
            response = await self.__connection.post(
                self.get_token_url(), data=payload, headers=headers
            )
            self.__parse_token_response(response.json())
        except httpx.HTTPStatusError as ex:
            except_errors = [
                "Refresh token expired",
                "Token is not active",
                "Session not active",
            ]
            error_description = ex.response.json().get("error_description", "")
            if ex.response.status_code == 400 and any(
                error in error_description for error in except_errors
            ):
                await self.__token()

    async def __token(self):
        headers = httpx.Headers({"Content-Type": "application/x-www-form-urlencoded"})
        payload = remove_none(
            {
                "client_id": self._client_id,
                "grant_type": self._grant_type,
                "client_secret": self._client_secret,
                "username": self._username,
                "password": self._password,
            }
        )
        response = await self.__connection.post(
            self.get_token_url(), data=payload, headers=headers
        )
        self.__parse_token_response(response.json())

    async def __get_connection(self):
        access_token = await self.get_access_token()

        def auth_interceptor(request: httpx.Request) -> httpx.Request:
            request.headers["Authorization"] = f"Bearer {access_token}"
            return request

        self.__connection.auth = auth_interceptor
        return self.__connection



import uuid
from dataclasses import asdict as _asdict
from typing import Any, Optional, TypeVar, Union

import httpx


def remove_none(dictionary: dict[str, Any]) -> dict[str, Any]:
    def recurse(value: Union[dict, Any]):
        if type(value) is dict:
            return remove_none(value)
        else:
            return value

    return {k: recurse(v) for k, v in dictionary.items() if v is not None}


def asdict(dataclass) -> dict[str, Any]:
    return remove_none(_asdict(dataclass))


def is_uuid(value: str) -> bool:
    try:
        uuid.UUID(value)
        return True
    except ValueError:
        return False


def get_resource_id_in_location_header(
    response: httpx.Response, is_no_uuid=False
) -> str:
    location_header = response.headers.get("location")
    if not location_header or type(location_header) != str:
        raise Exception("The response headers didn't include location.")
    resource_id = location_header.split("/").pop()
    is_correct_type = type(resource_id) == str and (is_no_uuid or is_uuid(resource_id))
    if not is_correct_type:
        raise Exception("Resource id couldn't be found in location header.")
    return resource_id


T = TypeVar("T")


def cast_non_optional(arg: Optional[T]) -> T:
    assert arg is not None
    return arg



from typing import Optional

from keycloak_admin_aio._lib.utils import remove_none
from keycloak_admin_aio.types.types import UserSession

from .... import KeycloakResource


class ClientsByIdUserSessions(KeycloakResource):
    """User sessions for clients by id.

    .. code:: python

        from keycloak_admin_aio import KeycloakAdmin, UserSession

        kc: KeycloakAdmin  # must be instantiated
        client_uuid: str
    """

    def get_url(self) -> str:
        return f"{self._get_parent_url()}/user-sessions"

    async def get(
        self, first: Optional[int] = None, max: Optional[int] = None
    ) -> list[UserSession]:
        """Get user sessions for a client.

        .. code:: python

            user_sessions: lists[UserSession] = await kc.clients.by_id(client_uuid).user_sessions.get()
        """
        connection = await self._get_connection()
        params = remove_none({"first": first, "max": max})
        response = await connection.get(self.get_url(), params=params)
        return UserSession.from_list(response.json())



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Literal, Optional

from ._data_class import DataClass


@dataclass
class RoleRepresentationComposites(DataClass):
    """https://www.keycloak.org/docs-api/15.0/rest-api/index.html#_rolerepresentation-composites"""

    realm: Optional[list[str]] = None
    client: Optional[dict[str, list[str]]] = None


@dataclass
class RoleRepresentation(DataClass):
    """https://www.keycloak.org/docs-api/15.0/rest-api/index.html#_rolerepresentation"""

    id: Optional[str] = None
    name: Optional[str] = None
    description: Optional[str] = None
    containerId: Optional[str] = None
    composites: Optional[RoleRepresentationComposites] = None
    composite: Optional[bool] = None
    clientRole: Optional[bool] = None
    attributes: Optional[dict[str, list[str]]] = None


@dataclass
class ClientMappingsRepresentation(DataClass):
    """https://www.keycloak.org/docs-api/15.0/rest-api/index.html#_clientmappingsrepresentation"""

    id: Optional[str] = None
    client: Optional[str] = None
    mappings: Optional[list[RoleRepresentation]] = None


@dataclass
class MappingsRepresentation(DataClass):
    """https://www.keycloak.org/docs-api/15.0/rest-api/index.html#_mappingsrepresentation"""

    clientMappings: Optional[dict[str, ClientMappingsRepresentation]] = None
    realmMappings: Optional[list[RoleRepresentation]] = None


@dataclass
class ProtocolMapperRepresentation(DataClass):
    """https://www.keycloak.org/docs-api/15.0/rest-api/index.html#_protocolmapperrepresentation"""

    id: Optional[str] = None
    name: Optional[str] = None
    protocol: Optional[str] = None
    protocolMapper: Optional[str] = None


@dataclass
class ClientScopeRepresentation(DataClass):
    """https://www.keycloak.org/docs-api/15.0/rest-api/index.html#_clientscoperepresentation"""

    id: Optional[str] = None
    name: Optional[str] = None
    description: Optional[str] = None
    protocol: Optional[str] = None
    protocolMappers: Optional[list[ProtocolMapperRepresentation]] = None
    attributes: Optional[dict[str, str]] = None


@dataclass
class FederatedIdentityRepresentation(DataClass):
    """https://www.keycloak.org/docs-api/15.0/rest-api/index.html#_federatedidentityrepresentation"""

    userId: Optional[str] = None
    userName: Optional[str] = None
    identityProvider: Optional[str] = None


@dataclass
class CredentialRepresentation(DataClass):
    """https://www.keycloak.org/docs-api/15.0/rest-api/index.html#_credentialrepresentation"""

    id: Optional[str] = None
    createdDate: Optional[int] = None
    credentialData: Optional[str] = None
    priority: Optional[int] = None
    secretData: Optional[str] = None
    temporary: Optional[bool] = None
    type: Optional[str] = None
    userLabel: Optional[str] = None
    value: Optional[str] = None


@dataclass
class UserConsentRepresentation(DataClass):
    """https://www.keycloak.org/docs-api/15.0/rest-api/index.html#_userconsentrepresentation"""

    clientId: Optional[str] = None
    createdDate: Optional[int] = None
    grantedClientScopes: Optional[list[str]] = None
    lastUpdatedDate: Optional[int] = None


@dataclass
class UserRepresentation(DataClass):  # type: ignore
    """https://www.keycloak.org/docs-api/15.0/rest-api/index.html#_userrepresentation"""

    id: Optional[str] = None
    firstName: Optional[str] = None
    lastName: Optional[str] = None
    email: Optional[str] = None
    emailVerified: Optional[bool] = None
    username: Optional[str] = None
    requiredActions: Optional[list[str]] = None
    createdTimestamp: Optional[int] = None
    federationLink: Optional[str] = None
    federatedIdentities: Optional[list[FederatedIdentityRepresentation]] = None
    enabled: Optional[bool] = None
    disableableCredentialTypes: Optional[list[str]] = None
    credentials: Optional[list[CredentialRepresentation]] = None
    notBefore: Optional[int] = None
    access: Optional[dict[str, bool]] = None
    attributes: Optional[dict[str, list[str]]] = None
    clientConsents: Optional[list[UserConsentRepresentation]] = None
    totp: Optional[bool] = None
    self: Optional[str] = None
    origin: Optional[str] = None
    realmRoles: Optional[list[str]] = None
    groups: Optional[list[str]] = None
    serviceAccountClientId: Optional[str] = None
    clientRoles: Optional[dict[str, Any]] = None


@dataclass
class GroupRepresentation(DataClass):
    """https://www.keycloak.org/docs-api/15.0/rest-api/index.html#_grouprepresentation"""

    id: Optional[str] = None
    name: Optional[str] = None
    realmRoles: Optional[list[str]] = None
    path: Optional[str] = None
    subGroups: Optional[list[GroupRepresentation]] = None
    clientRoles: Optional[dict[str, Any]] = None
    attributes: Optional[dict[str, list[str]]] = None
    access: Optional[dict[str, bool]] = None


@dataclass
class ScopeRepresentation(DataClass):
    """https://www.keycloak.org/docs-api/15.0/rest-api/index.html#_scoperepresentation"""

    id: Optional[str] = None
    name: Optional[str] = None
    displayName: Optional[str] = None
    iconUri: Optional[str] = None
    policies: Optional[list[PolicyRepresentation]] = None
    resources: Optional[list[ResourceRepresentation]] = None


@dataclass
class ResourceRepresentation(DataClass):
    """https://www.keycloak.org/docs-api/15.0/rest-api/index.html#_resourcerepresentation"""

    id: Optional[str] = None
    name: Optional[str] = None
    displayName: Optional[str] = None
    icon_uri: Optional[str] = None
    attributes: Optional[dict[str, str]] = None
    ownerManagedAccess: Optional[bool] = None
    scopes: Optional[list[ScopeRepresentation]] = None
    type: Optional[str] = None
    uris: Optional[list[str]] = None


@dataclass
class PolicyRepresentation(DataClass):
    """https://www.keycloak.org/docs-api/15.0/rest-api/index.html#_policyrepresentation"""

    id: Optional[str] = None
    name: Optional[str] = None
    description: Optional[str] = None
    config: Optional[dict[str, str]] = None
    owner: Optional[str] = None
    logic: Optional[str] = None
    policies: Optional[list[str]] = None
    resources: Optional[list[str]] = None
    resourcesData: Optional[list[ResourceRepresentation]] = None
    scopes: Optional[list[str]] = None
    scopesData: Optional[list[ScopeRepresentation]] = None
    type: Optional[str] = None


@dataclass
class ResourceServerRepresentation(DataClass):
    """https://www.keycloak.org/docs-api/15.0/rest-api/index.html#_resourceserverrepresentation"""

    id: Optional[str] = None
    name: Optional[str] = None
    allowRemoteResourceManagement: Optional[bool] = None
    clientId: Optional[str] = None
    decisionStrategy: Optional[Literal["AFFIRMATIVE", "UNANIMOUS", "CONSENSUS"]] = None
    policyEnforcementMode: Optional[
        Literal["ENFORCING", "PERMISSIVE", "DISABLED"]
    ] = None
    resources: Optional[list[ResourceRepresentation]] = None
    scopes: Optional[list[ScopeRepresentation]] = None
    policies: Optional[list[PolicyRepresentation]] = None


@dataclass
class ClientRepresentation(DataClass):
    """https://www.keycloak.org/docs-api/15.0/rest-api/index.html#_clientrepresentation"""

    id: Optional[str] = None
    name: Optional[str] = None
    access: Optional[dict[str, bool]] = None
    adminUrl: Optional[str] = None
    alwaysDisplayInConsole: Optional[bool] = None
    attributes: Optional[dict[str, str]] = None
    authenticationFlowBindingOverrides: Optional[dict[str, str]] = None
    authorizationServicesEnabled: Optional[bool] = None
    authorizationSettings: Optional[ResourceServerRepresentation] = None
    baseUrl: Optional[str] = None
    bearerOnly: Optional[bool] = None
    clientAuthenticatorType: Optional[str] = None
    clientId: Optional[str] = None
    consentRequired: Optional[bool] = None
    defaultClientScopes: Optional[list[str]] = None
    description: Optional[str] = None
    directAccessGrantsEnabled: Optional[bool] = None
    enabled: Optional[bool] = None
    frontchannelLogout: Optional[bool] = None
    fullScopeAllowed: Optional[bool] = None
    implicitFlowEnabled: Optional[bool] = None
    nodeReRegistrationTimeout: Optional[int] = None
    notBefore: Optional[int] = None
    oauth2DeviceAuthorizationGrantEnabled: Optional[bool] = None
    optionalClientScopes: Optional[list[str]] = None
    origin: Optional[str] = None
    protocol: Optional[str] = None
    protocolMappers: Optional[list[ProtocolMapperRepresentation]] = None
    publicClient: Optional[bool] = None
    redirectUris: Optional[list[str]] = None
    registeredNodes: Optional[dict[str, str]] = None
    registrationAccessToken: Optional[str] = None
    rootUrl: Optional[str] = None
    secret: Optional[str] = None
    serviceAccountsEnabled: Optional[bool] = None
    standardFlowEnabled: Optional[bool] = None
    surrogateAuthRequired: Optional[bool] = None
    webOrigins: Optional[list[str]] = None


@dataclass
class RequiredActionProviderRepresentation(DataClass):
    """https://www.keycloak.org/docs-api/15.0/rest-api/index.html#_requiredactionproviderrepresentation"""

    alias: Optional[str] = None
    config: Optional[dict[str, Any]] = None
    defaultAction: Optional[bool] = None
    enabled: Optional[bool] = None
    name: Optional[str] = None
    priority: Optional[int] = None
    providerId: Optional[str] = None


@dataclass
class UserSession(DataClass):
    """This is no default Keycloak representation."""

    clients: Optional[dict[str, str]] = None
    id: Optional[str] = None
    ipAddress: Optional[str] = None
    lastAccess: Optional[int] = None
    start: Optional[int] = None
    userId: Optional[str] = None
    username: Optional[str] = None


ResourceType = Literal[
    "REALM",
    "REALM_ROLE",
    "REALM_ROLE_MAPPING",
    "REALM_SCOPE_MAPPING",
    "AUTH_FLOW",
    "AUTH_EXECUTION_FLOW",
    "AUTH_EXECUTION",
    "AUTHENTICATOR_CONFIG",
    "REQUIRED_ACTION",
    "IDENTITY_PROVIDER",
    "IDENTITY_PROVIDER_MAPPER",
    "PROTOCOL_MAPPER",
    "USER",
    "USER_LOGIN_FAILURE",
    "USER_SESSION",
    "USER_FEDERATION_PROVIDER",
    "USER_FEDERATION_MAPPER",
    "GROUP",
    "GROUP_MEMBERSHIP",
    "CLIENT",
    "CLIENT_INITIAL_ACCESS_MODEL",
    "CLIENT_ROLE",
    "CLIENT_ROLE_MAPPING",
    "CLIENT_SCOPE",
    "CLIENT_SCOPE_MAPPING",
    "CLIENT_SCOPE_CLIENT_MAPPING",
    "CLUSTER_NODE",
    "COMPONENT",
    "AUTHORIZATION_RESOURCE_SERVER",
    "AUTHORIZATION_RESOURCE",
    "AUTHORIZATION_SCOPE",
    "AUTHORIZATION_POLICY",
    "CUSTOM",
]

OperationType = Literal["CREATE", "DELETE", "ACTION", "UPDATE"]



from keycloak_admin_aio.types import ClientScopeRepresentation

from .... import (
    AttachedResources,
    KeycloakResource,
    KeycloakResourceWithIdentifierGetter,
)
from .by_id import ClientsByIdDefaultClientScopesById


class ClientsByIdDefaultClientScopes(KeycloakResource):
    """Default client scopes for clients by UUID.

    .. code:: python

        from keycloak_admin_aio import KeycloakAdmin, ClientScopeRepresentation

        kc: KeycloakAdmin  # must be instantiated
        client_uuid: str
    """

    _keycloak_resources: AttachedResources = [
        ("by_id", ClientsByIdDefaultClientScopesById)
    ]
    by_id: KeycloakResourceWithIdentifierGetter[ClientsByIdDefaultClientScopesById]

    def get_url(self) -> str:
        return f"{self._get_parent_url()}/default-client-scopes"

    async def get(self) -> list[ClientScopeRepresentation]:
        """Get default clients scopes for a client by UUID.

        .. code:: python

            client_scope_resource = kc.clients.by_id(client_uuid)
            client_scope_representations: list[
                ClientScopeRepresentation
            ] = await client_scope_resource.default_client_scopes.get()
        """
        connection = await self._get_connection()
        response = await connection.get(self.get_url())
        return ClientScopeRepresentation.from_list(response.json())



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['keycloak_admin_aio',
 'keycloak_admin_aio._resources',
 'keycloak_admin_aio._resources.admin_events',
 'keycloak_admin_aio._resources.authentication',
 'keycloak_admin_aio._resources.authentication.required_actions',
 'keycloak_admin_aio._resources.client_scopes',
 'keycloak_admin_aio._resources.client_scopes.by_id',
 'keycloak_admin_aio._resources.client_scopes.by_id.scope_mappings',
 'keycloak_admin_aio._resources.client_scopes.by_id.scope_mappings.realm',
 'keycloak_admin_aio._resources.clients',
 'keycloak_admin_aio._resources.clients.by_id',
 'keycloak_admin_aio._resources.clients.by_id.default_client_scopes',
 'keycloak_admin_aio._resources.clients.by_id.default_client_scopes.by_id',
 'keycloak_admin_aio._resources.clients.by_id.user_sessions',
 'keycloak_admin_aio._resources.groups',
 'keycloak_admin_aio._resources.groups.by_id',
 'keycloak_admin_aio._resources.groups.by_id.children',
 'keycloak_admin_aio._resources.groups.by_id.members',
 'keycloak_admin_aio._resources.roles',
 'keycloak_admin_aio._resources.roles.by_id',
 'keycloak_admin_aio._resources.roles.by_id.composites',
 'keycloak_admin_aio._resources.roles.by_name',
 'keycloak_admin_aio._resources.roles.by_name.composites',
 'keycloak_admin_aio._resources.sessions',
 'keycloak_admin_aio._resources.sessions.by_id',
 'keycloak_admin_aio._resources.users',
 'keycloak_admin_aio._resources.users.by_id',
 'keycloak_admin_aio._resources.users.by_id.execute_actions_email',
 'keycloak_admin_aio._resources.users.by_id.groups',
 'keycloak_admin_aio._resources.users.by_id.groups.by_id',
 'keycloak_admin_aio._resources.users.by_id.role_mappings',
 'keycloak_admin_aio._resources.users.by_id.role_mappings.realm',
 'keycloak_admin_aio.types']

package_data = \
{'': ['*'], 'keycloak_admin_aio': ['_lib/*']}

install_requires = \
['dacite>=1.6.0,<2.0.0', 'httpx>=0.21.0,<0.22.0']

setup_kwargs = {
    'name': 'keycloak-admin-aio',
    'version': '1.2.2',
    'description': 'async keycloak admin api wrapper',
    'long_description': 'What is keycloak_admin_aio?\n---------------------------\n\nThis package provides an asynchronous api wrapper for the `keycloak admin api\n<https://www.keycloak.org/docs-api/15.0/rest-api>`_.\n\nThe main dependencies are:\n\n- `httpx <https://github.com/encode/httpx/>`_: asynchronous http client\n- `dacite <https://github.com/konradhalas/dacite>`_: parse nested dictionaries into nested dataclasses\n\nLinks:\n\n- `Source code <https://github.com/delphai/keycloak-admin-aio>`_\n- `Documentation <https://delphai.github.io/keycloak-admin-aio/>`_\n- `Pypi <https://pypi.org/project/keycloak-admin-aio/>`_\n\nHow to install?\n---------------\n\n.. code:: shell\n\n   poetry add keycloak-admin-aio\n\nHow to use it?\n--------------\n\n.. code:: python\n\n    import asyncio\n    from keycloak_admin_aio import KeycloakAdmin\n\n    server_url = "http://localhost:8080/auth"\n    client_id = "admin-cli"  # used by default\n    realm = "master"  # used by default\n\nWith administrator username and password:\n\n.. code:: python\n\n    keycloak_admin_args = {\n        "server_url": server_url,\n        "client_id": client_id,\n        "realm": realm,\n        "username": "admin",\n        "password": "password123",\n    }\n\n    async def main():\n        async with KeycloakAdmin.with_password(**keycloak_admin_args) as kc:\n            users = await kc.users.get(email="@google")\n            await asyncio.gather(\n                *[\n                    kc.users.by_id(user.id).execute_actions_email.send_email(\n                        ["UPDATE_PASSWORD"]\n                    )\n                    for user in users\n                ]\n            )\n\n    asyncio.run(main())\n\nWith client credentials:\n\n.. code:: python\n\n    keycloak_admin_args = {\n        "realm": realm,\n        "server_url": server_url,\n        "client_id": client_id,\n        "client_secret": "the_secret",\n    }\n\n    async def main():\n        async with KeycloakAdmin.with_client_credentials(**keycloak_admin_args) as kc:\n            ...\n\n    asyncio.run(main())\n\nLicense\n-------\n\n`Apache License, Version 2.0 <https://www.apache.org/licenses/LICENSE-2.0>`_\n\n\n© Copyright 2021, delphai by AtomLeap GmbH\n',
    'author': 'Nicklas Sedlock',
    'author_email': 'nicklas@delphai.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/delphai/keycloak-admin-aio',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



"""https://www.keycloak.org/docs-api/15.0/rest-api/index.html#_clients_resource

One should not be confused between ``client_id`` and the UUID a ``client`` has.
Therefor this section will refer to variable names containing the UUID of a
client as ``client_uuid``.
"""

from .clients import Clients



"""This package provides classes for Keycloak resources."""

from .keycloak_resource import (
    KeycloakResource,
    AttachedResources,
    KeycloakResourceWithIdentifier,
    KeycloakResourceWithIdentifierGetter,
)

from .admin_events import AdminEvents
from .authentication import Authentication
from .client_scopes import ClientScopes
from .clients import Clients
from .groups import Groups
from .roles import Roles
from .users import Users
from .sessions import Sessions



from keycloak_admin_aio.types import UserRepresentation

from ... import AttachedResources, KeycloakResourceWithIdentifier
from .execute_actions_email import UsersByIdExecuteActionsEmail
from .groups import UsersByIdGroups
from .role_mappings import UsersByIdRoleMappings


class UsersById(KeycloakResourceWithIdentifier):
    """Provides the Keycloak users by id resource.

    .. code:: python

        from keycloak_admin_aio import KeycloakAdmin, UserRepresentation

        kc: KeycloakAdmin  # needs to be instantiated
        user_id: str  # uuid
    """

    _keycloak_resources: AttachedResources = [
        ("role_mappings", UsersByIdRoleMappings),
        ("groups", UsersByIdGroups),
        ("execute_actions_email", UsersByIdExecuteActionsEmail),
    ]
    role_mappings: UsersByIdRoleMappings
    groups: UsersByIdGroups
    execute_actions_email: UsersByIdExecuteActionsEmail

    def get_url(self) -> str:
        return f"{self._get_parent_url()}/{self.identifier}"

    async def get(self) -> UserRepresentation:
        """Get user by id.

        .. code:: python

            user: UserRepresentation = await kc.users.by_id(user_id).get()
        """
        connection = await self._get_connection()
        response = await connection.get(self.get_url())
        return UserRepresentation.from_dict(response.json())

    async def update(self, user_representation: UserRepresentation):
        """Update user by id.

        .. code:: python

            user_representation = UserRepresentation(first_name="New name")
            await kc.users.by_id(user_id).update(user_representation)
        """
        connection = await self._get_connection()
        await connection.put(self.get_url(), json=user_representation.to_dict())

    async def delete(self):
        """Delete user by id.

        .. code:: python

            await kc.users.by_id(user_id).delete()
        """
        connection = await self._get_connection()
        await connection.delete(self.get_url())



from keycloak_admin_aio.types import GroupRepresentation

from ... import AttachedResources, KeycloakResourceWithIdentifier
from .children import GroupsByIdChildren
from .members import GroupsByIdMembers


class GroupsById(KeycloakResourceWithIdentifier):
    """Groups by id.

    .. code:: python

        from keycloak_admin_aio import KeycloakAdmin, GroupRepresentation

        kc: KeycloakAdmin  # must be instantiated
        group_id: str  # uuid
    """

    _keycloak_resources: AttachedResources = [
        ("children", GroupsByIdChildren),
        ("members", GroupsByIdMembers),
    ]
    children: GroupsByIdChildren
    members: GroupsByIdMembers

    def get_url(self) -> str:
        return f"{self._get_parent_url()}/{self.identifier}"

    async def get(self) -> GroupRepresentation:
        """Get a group by id.

        .. code:: python

            group: GroupRepresentation = await kc.groups.by_id(group_id).get()
        """
        connection = await self._get_connection()
        response = await connection.get(self.get_url())
        return GroupRepresentation.from_dict(response.json())

    async def update(self, group_representation: GroupRepresentation):
        """Update a group by id.

        .. code:: python

            group_representation = GroupRepresentation(name="group-name")
            await kc.groups.by_id(group_id).update(group_representation)
        """
        connection = await self._get_connection()
        await connection.put(self.get_url(), json=group_representation.to_dict())

    async def delete(self):
        """Delete a group by id.

        .. code:: python

            await kc.groups.by_id(group_id).delete()
        """
        connection = await self._get_connection()
        await connection.delete(self.get_url())
